//////////////////////////////////////////////////////////////////////////////
//
//		Computing Fano 3-folds with index f
//
//  Gavin Brown, Kaori Suzuki, February 2006, August 2010.
//
//  Having started magma in the same directory as this file, write
/*
AttachSpec("fano3.spec");
*/
// or attach the files separately by
/*
Attach("fanosuzuki.m");
Attach("fano3folds.m");
Attach("fanobaskets.m");
*/
//
// To compute all (resp all Bogomolov--Kawamata stable) Fano 3-folds do:
//
/*
To run for index >= 3
=====================

allB := [ FanoBaskets(24,f : Stable:=false) : f in [3..19] ];
stableB := [ FanoBaskets(24,f) : f in [3..19] ];

[ #c : c in allB ];                                          
	// [ 231, 124, 63, 11, 23, 10, 2, 1, 3, 0, 2, 0, 0, 0, 1, 0, 1 ]
[ #c : c in stableB ];
	// [ 181, 82, 34, 6, 12, 4, 2, 0, 3, 0, 2, 0, 0, 0, 1, 0, 1 ]

allfanos := [ [ Fano3fold(f,b) : b in allB[f-2] ] : f in [3..19] ];
stablefanos := [ [ Fano3fold(f,b) : b in allB[f-2] ] : f in [3..19] ];


To run for index 2
==================

Bs,Gs := FanoBaskets(24,2);	// about 2s
#Bs;	// 709
#&cat Gs;	// 1413
stablefano2s := [ Fano3fold(2,Bs[i],g) : g in Gs[i], i in [1..#Bs] ];	// ~3s
#stablefano2s;	// 1413

// or 

Bs,Gs := FanoBaskets(24,2 : Stable:=false);	// about 2s
#Bs;	// 721
#&cat Gs;	//1492
allfano2s := [ Fano3fold(2,Bs[i],g) : g in Gs[i], i in [1..#Bs] ];	// ~2s
#allfano2s;	// 1492


// But this has no weights info computed yet.

#[ X : X in fanos | Codimension(X) eq 1 ];	// about 140s

// Alternatively, to see the codims:

{* Codimension(X) : X in fanos *};

{* 1^^6, 2^^25, 3^^3, 4^^54, 5^^19, 6^^105, 7^^36, 8^^103, 9^^71, 10^^157, 
11^^99, 12^^98, 13^^105, 14^^92, 15^^87, 16^^110, 17^^95, 18^^113, 19^^94, 
20^^91, 21^^77, 22^^67, 23^^25, 24^^15, 25^^4, 26^^4, 27, 28, 29, 31^^2, 32, 33,
34^^2, 37, 41, 43, 53, 54^^2, 60, 61, 68, 70, 119 *}

// Or slowly:

for i in [1..#Bs ] do
	b := Bs[i];
	for g in Gs[i] do
		X := Fano2(b,g);
		X;
	end for;
end for; 

To run for index 1
==================
Bs,Gs := FanoBaskets(24,1);	// about 50s
#Bs;	// 7492
#&cat Gs;	// 39558

Bus,Gus := FanoBaskets(24,1 : Stable:=false);	// about 55s
#Bus;	// 7683
#&cat Gus;	// 52654

*/

import "fanosuzuki.m":
	fano_Ac,
	fano_hilbert_series,
	fano_hilbert_series_g,
	fano_degree_g;

//////////////////////////////////////////////////////////////////////////////
//  Housekeeping functions --- data types etc.
//////////////////////////////////////////////////////////////////////////////

forward base_genus;

intrinsic Fano3fold(f::RngIntElt,B::GRBskt,g::RngIntElt) -> GRFano
{A Fano 3-fold of Fano index f = 1 or 2 and with basket of singularities B
and Fano genus g >= 0}
    require f in {1,2}: "Genus argument only needed when Fano index f = 1 or 2";
    require IsTerminalThreefold(B):
        "Basket must contain only terminal 3-fold singularities";
    require &and[ (f mod Index(p)) eq TerminalPolarisation(p)[1] :
        p in Points(B) ]: "Singularities are not properly polarised "
            * "for the given Fano index f =",f;
    g_min := base_genus(f,B);
    require g ge g_min: "Third argument g must be >=",g_min;
    // make the object X and attach all the Fano data.
    X := HackobjCreateRaw(GRFano);
    X`fanoindex := f;
    X`fanogenus := g;
    X`fanobasegenus := g_min;
    X`basket := B;
    X`dim := 3;
    h,d,a := fano_hilbert_series_g(g,f,B);
    require d gt 0 : "Data determines a degree <= 0";
    X`hilbert := h;
    X`degree := d;
    X`Ac := a;
    return X;
end intrinsic;

intrinsic Fano3fold(f::RngIntElt,B::GRBskt) -> GRFano
{A Fano 3-fold of Fano index f >= 3 and with basket of singularities B}
    require f ge 3: "Genus argument needed when Fano index f >= 3";
    require IsTerminalThreefold(B):
        "Basket must contain only terminal 3-fold singularities";
    require &and[ (f mod Index(p)) eq TerminalPolarisation(p)[1] :
        p in Points(B) ]: "Singularities are not properly polarised "
            * "for the given Fano index f =",f;
    // make the object X and attach all the Fano data.
    X := HackobjCreateRaw(GRFano);
    X`fanoindex := f;
    X`basket := B;
    X`dim := 3;
    h,d,a := fano_hilbert_series(f,B);
    require d gt 0 : "Data determines a degree <= 0";
    X`hilbert := h;
    X`degree := d;
    X`Ac := a;
    return X;
end intrinsic;


/////////////////////////////////////////////////////////////////////
//		Genus bounds when f = 1 or 2
/////////////////////////////////////////////////////////////////////

// The Bogomolov--Kawamata stability bound
// 'stable' is a boolean: true imposes the stronger bounds.
function bogomolov(f,g,B,stable)
    if stable then
	return f^2*fano_degree_g(g,f,B) - 3*12*fano_Ac(f,B);
    else
	return (4*f^2-3*f)*fano_degree_g(g,f,B) - 4*f*12*fano_Ac(f,B);
    end if;
end function;

function max_genus(f,B : Stable:=true)
    if not f in {1,2} then
        return 0;   // this value is irrelevant
    end if;
    g := 0;	// starting value - could be anything.
    if bogomolov(f,g,B,Stable) gt 0 then
      repeat
          g -:= 1;
      until bogomolov(f,g,B,Stable) le 0;
    else
      repeat
          glast := g;
          g +:= 1;
      until bogomolov(f,g,B,Stable) gt 0;
      g := glast;
    end if;
    return g;
end function;

// The minimum value of the Fano genus g for which the triple (f,B,g)
// determine a positive Fano degree.
function base_genus(f,B)
    if not f in {1,2} then
        return 0;   // this value is irrelevant
    end if;
    // I could be smarter about this since the degree is linear in g.
    g := 0;
    if fano_degree_g(g,f,B) le 0 then
      repeat
          g +:= 1;
      until fano_degree_g(g,f,B) gt 0;
    else
      repeat
          glast := g;
          g -:= 1;
      until fano_degree_g(g,f,B) le 0;
      g := glast;
    end if;
    return g;
end function;

