////////////////////////////////////////////////////////////////////////////////
// Making the baskets of Fano 3-folds of Fano index f with 'bound' n
//
// Idea: first make a big collection of baskets satisfying the weaker
// conditions: N <= 15 and sum(r-1) < 24, where N = #singular points.
// NB n == 24 here.
// Then test the real conditions
//	N <= 15 and (3/2)N <= sum(r - (1/r)) < 24
// on each of these.
// When f = 1 or 2 the parameter g also plays a role: it appears in formulas
// for A^3 and for the Hilbert coefficients.
//
// We refer to sum (r - 1/r) as the 'Kawamata sum'.
//
////////////////////////////////////////////////////////////////////////////////

import "fanosuzuki.m":
	fano_Ac,
	fano_degree,
	fano_degree_g,
	contribution;

import "fano3folds.m":
	base_genus,
	max_genus,
	bogomolov;


/////////////////////////////////////////////////////////////////////
// This is the main intrinsic.
// Conditions are imposed separately by procedures written below.
/////////////////////////////////////////////////////////////////////

function coefficient(f,B,n)
    return 1 + (1/12) * fano_degree(f,B) * n*(n+f)*(2*n+f)
        + n * fano_Ac(f,B)
        + &+[Rationals()| contribution(f,p,n mod Index(p)) :
                                p in Points(B) ];
end function;


function coefficient_g(g,f,B,n) // the g in fano_degree is the only difference
    return 1 + (1/12) * fano_degree_g(g,f,B) * n*(n+f)*(2*n+f)
        + n * fano_Ac(f,B)
        + &+[Rationals()| contribution(f,p,(n mod Index(p))) :
                                p in Points(B) ];
end function;

intrinsic FanoBaskets(n::RngIntElt,f::RngIntElt :
		Order:="default",Stable:=true) -> SeqEnum, SeqEnum
{All baskets of singularities appearing on Fano 3-folds with Fano index f
and Kawamata index sum at most n. If f = 1 or 2, then a parallel sequence
of possible genus values g is also returned.}

	// Cook up all baskets with the Kawamata sum bounded by n.
	B := FanoIsolatedBaskets(n,f);

	if f in {1,2} then
		// For each b in B find the values of g = g_min..g_kawamata
		// for which A^3 > 0 and Kawamata condition is satisfied.
		g_mins := [ base_genus(f,b) : b in B ];
		g_maxs := [ max_genus( f, B[i] : Stable:=Stable) :i in [1..#B]];
		// For each b,g pair, check the negative coefficients.
		newB := [];
		G := [ Parent([1,2,3]) | ];
		for i in [1..#B] do
			b := B[i];
			gg := [ Integers() | ];
			g_min := g_mins[i];
			g_max := g_maxs[i];
			// Impose vanishing of coeffs of t^-1, t^-2,...
			for g in [g_min..g_max] do
				if &and [ coefficient_g(g,f,b,n) eq 0 :
						n in [-Round(f-1)..-1] ] then
					Append(~gg, g);
				end if;
			end for;
			if #gg gt 0 then
				Append(~newB,b);
				Append(~G,gg);
			end if;
		end for;
		B := newB;
		require #B eq #G: "Something is badly wrong";
		return B,G;
	else
		// Impose A^3 > 0 on the baskets:
#B;
		B := [ b : b in B | fano_degree(f,b) gt 0 ];
#B;
		// Impose vanishing of coeffs of t^-1, t^-2,...:
		B := [ b : b in B |
			&and[coefficient(f,b,n) eq 0 : n in [-Round(f-1)..-1]]];
#B;
		// Impose stability if requested:
		B := [ b : b in B | bogomolov(f,0,b,Stable) le 0 ];
#B;
		return B,_;
	end if;
end intrinsic;


///////////////////////////////////////////////////////////////////////
//  Making the raw list of isolated singularities with given index f
//
// This is purely combinatorial - no RR involved.
//
///////////////////////////////////////////////////////////////////////

forward indices, small_coprimes, round_up;
intrinsic FanoIsolatedBaskets(n::RngIntElt,f::RngIntElt) -> SeqEnum
{Large list of possible baskets of isolated singularities appearing
on Fano 3-folds with Fano index f and Kawamata index sum at most n}

    // Idea: first make a big collection of baskets satisfying the weaker
    // conditions: N <= 15 and sum(r-1) < 24, where N = #singular points.
    // NB n == 24 here.

    tt := Cputime();

    vprintf User1: "Building collections of singularities...\n";
    // Precompute all sequences of singularity types [r,a] that might be needed.
    // The result is called 'sing_combs'.
    max_num_sings_per_r := [ round_up(n/(r-1)) - 1 : r in [2..n] ];
    A := [ small_coprimes(i) : i in [2..n] ];
    // Entry 'poss_as_per_r[r-1][j]' is a sequence containing sequences of
    // 'a's (in notation [r,a]) that could appear in a collection of j sings
    // of index r.
    poss_as_per_r := [];
    for r in [2..n]  do
	poss_a := [];
	for j in [1..max_num_sings_per_r[r-1]] do
	    poss_a[j] := Sort(SetToSequence({ Sort(S) :
				S in Subsequences(A[r-1],j) }));
	end for;
	poss_as_per_r[r-1] := poss_a;
    end for;
    // Add the index 'r' to each of the elements of 'poss_as_per_r' to make
    // baskets.
    sing_combs := [];
    for r in [2..n] do
	poss_comb := [];
	for j in [1..max_num_sings_per_r[r-1]] do
	    papj := poss_as_per_r[r-1][j];
	    baskets := [];
	    for k in [1..#papj] do
		Append(~baskets,[ [r,i] : i in papj[k]]);
	    end for;
	    Append(~poss_comb,baskets);
	end for;
	sing_combs[r-1] := poss_comb;
    end for;

    // Compute prima facie allowable collections of indices according to the
    // coarse formula \sum (ri - 1) < n. Redescribe collections like
    // [ 2,2,2,2,2,3,4,4 ] as [ [2,5],[3,1],[4,2] ].
    // Result is called 'Ifactors'.
    I0 := indices(n);
    I := [ R : R in I0 |
			#R le 15 and
			(3/2)*#R le fanosum and fanosum lt 31
		where fanosum is &+[Rationals()| r - 1/r : r in R ] ];
    Ifactors := [];
    for i in I do
	new_term := [];
	for r in [2..n] do
	    k := #[ z : z in i | z eq r ];
	    if k gt 0 then
		Append(~new_term,[r,k]);
	    end if;
	end for;
	Append(~Ifactors,new_term);
    end for;

    // Translate the result into sequences of baskets.
    result := [];
    for X in Ifactors do
	Xbaskets := [ [] ];
	for inds in X do
	    poss_sings := sing_combs[inds[1]-1,inds[2]];
	    Xbaskets_temp := [];
	    for S in poss_sings do
		for B in Xbaskets do
		    Append(~Xbaskets_temp,B cat S);
		end for;
	    end for;
	    Xbaskets := Xbaskets_temp;
	end for;
	result cat:= Xbaskets;
    end for;
    vprintf User1: "Made %o raw baskets\t",#result;
    vprintf User1: "\tTime: %o\n",Cputime(tt);

    // don't include nonisolated sings: f must be coprime to index p[1]
    vprintf User1: "Assembling only those with isolated singularities...\n";
    result1 := [ B : B in result | &and[ GCD(p[1],f) eq 1 : p in B ] ];

    // Turn this raw result into singularities before making subtle tests.
    BB := [ Basket([Point(p[1],[p[2],p[1]-p[2],f]): p in B ]): B in result1 ];
    vprintf User1: "Now have %o baskets\t",#BB;

    return BB;
end intrinsic;

///////////////////////////////////////////////////////////////////////

forward add_term, sub_sum;
function indices(n)
    // return sequences of indices [ r1,r2,... ] so that \sum (ri - 1) < n.
    I := [ [ Integers() | ] ];
    Itemp := I;
    repeat
	S := &cat [add_term(B,n) : B in Itemp ];
	I cat:= S;
	Itemp := [ B : B in S | sub_sum(B) lt n ];
    until #Itemp eq 0;
    return I;
end function;

function add_term(B,n)
    new := [];
    if #B eq 0 then
	m0 := 2;
	m1 := n;
    else
	m0 := B[#B];
	m1 := n - sub_sum(B);
    end if;
    for i in [m0..m1] do
	C := B;
	Append(~C,i);
	Append(~new,C);
    end for;
    return new;
end function;

function sub_sum(B)
    return &+ [ b - 1 : b in B ];
end function;

function small_coprimes(r)
    n := IsEven(r) select Integers() ! (r/2) else Integers() ! ((r-1)/2);
    return { a : a in [1..n] | GCD(r,a) eq 1 };
end function;

function round_up(q)
    n := Round(q);
    if n - q ge 0 then
	return n;
    else
	return n + 1;
    end if;
end function;

